from typing import Optional, Tuple

from nextline.spawned import OnEndTrace, OnStartTrace
from nextline.spec import hookimpl
from nextline.types import RunNo, TraceNo
from nextline.utils.pubsub.broker import PubSub


class TraceNumbersRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None
        self._trace_nos: Tuple[TraceNo, ...] = ()

    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_initialize_run(self, run_no: RunNo) -> None:
        self._run_no = run_no
        self._trace_nos = ()

    @hookimpl
    async def on_end_run(self) -> None:
        self._run_no = None

        self._trace_nos = ()
        await self._registry.publish('trace_nos', self._trace_nos)

    @hookimpl
    async def on_start_trace(self, event: OnStartTrace) -> None:
        trace_no = event.trace_no
        self._trace_nos = self._trace_nos + (trace_no,)
        await self._registry.publish('trace_nos', self._trace_nos)

    @hookimpl
    async def on_end_trace(self, event: OnEndTrace) -> None:
        trace_no = event.trace_no
        nosl = list(self._trace_nos)
        try:
            nosl.remove(trace_no)
        except ValueError:
            return
        self._trace_nos = tuple(nosl)
        await self._registry.publish('trace_nos', self._trace_nos)



from typing import Optional

from nextline.spawned import OnWriteStdout
from nextline.spec import hookimpl
from nextline.types import RunNo, StdoutInfo
from nextline.utils.pubsub.broker import PubSub


class StdoutRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None

    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_initialize_run(self, run_no: RunNo) -> None:
        self._run_no = run_no
        self._trace_nos = ()

    @hookimpl
    async def on_end_run(self) -> None:
        self._run_no = None

    @hookimpl
    async def on_write_stdout(self, event: OnWriteStdout) -> None:
        assert self._run_no is not None
        stdout_info = StdoutInfo(
            run_no=self._run_no,
            trace_no=event.trace_no,
            text=event.text,
            written_at=event.written_at,
        )
        await self._registry.publish('stdout', stdout_info)



__all__ = [
    'FilerByModule',
    'FilterByModuleName',
    'FilterLambda',
    'GlobalTraceFunc',
    'TaskAndThreadKeeper',
    'TaskOrThreadToTraceMapper',
    'LocalTraceFunc',
    'TraceCallHandler',
    'PdbInstanceFactory',
    'Prompt',
    'PeekStdout',
    'Repeater',
]


from .filter import FilerByModule, FilterByModuleName, FilterLambda
from .global_ import GlobalTraceFunc, TaskAndThreadKeeper, TaskOrThreadToTraceMapper
from .local_ import LocalTraceFunc, TraceCallHandler
from .pdb_ import PdbInstanceFactory, Prompt
from .peek import PeekStdout
from .repeat import Repeater



from __future__ import annotations

import threading
from asyncio import Task
from functools import lru_cache, partial
from logging import getLogger
from threading import Thread
from typing import Generator, Iterable, Iterator, Optional, Set

from apluggy import PluginManager, contextmanager

from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.plugin.types import TraceArgs
from nextline.utils import current_task_or_thread, match_any


class FilterByModuleName:
    '''Skip Python modules with names that match any of the patterns.'''

    @hookimpl
    def init(self, modules_to_skip: Iterable[str]):
        self._patterns = frozenset(modules_to_skip)

        # NOTE: Use lru_cache() as match_any() is slow
        self._match_any_ = lru_cache(partial(match_any, patterns=modules_to_skip))

    @hookimpl
    def filter(self, trace_args: TraceArgs) -> bool | None:
        frame = trace_args[0]
        module_name = frame.f_globals.get('__name__')
        matched = self._match_any_(module_name)
        return matched or None


class FilterLambda:
    '''Skip lambda functions.'''

    @hookimpl
    def filter(self, trace_args: TraceArgs) -> bool | None:
        frame = trace_args[0]
        func_name = frame.f_code.co_name
        return func_name == '<lambda>' or None


class FilerByModule:
    '''Accept the first module and modules ever in the cmdloop() context.'''

    def __init__(self) -> None:
        self._modules_to_trace: Set[str] = set()
        self._first_module_added = False
        self._entering_thread: Optional[Thread] = None
        self._traced_tasks_and_threads: Set[Task | Thread] = set()
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    @contextmanager
    def context(self) -> Iterator[None]:
        self._entering_thread = threading.current_thread()
        msg = f'{self.__class__.__name__}: entering thread {self._entering_thread}'
        self._logger.info(msg)
        yield

    @hookimpl(trylast=True)
    def filter(self, trace_args: TraceArgs) -> bool | None:

        # Accept the first module in the main thread.
        if not self._first_module_added:
            if self._entering_thread == threading.current_thread():
                self._add(trace_args)
                self._first_module_added = True

        # Accept tasks and threads once they are accepted.
        task_or_thread = current_task_or_thread()
        if task_or_thread in self._traced_tasks_and_threads:
            return None

        # Accept if the module was in the cmdloop() in other tasks or threads.
        if self._to_trace(trace_args):
            self._traced_tasks_and_threads.add(task_or_thread)
            return None

        # Reject else.
        return True

    @hookimpl
    @contextmanager
    def on_cmdloop(self) -> Generator[None, str, None]:
        trace_args = self._hook.hook.current_trace_args()
        self._add(trace_args)
        yield

    def _add(self, trace_args: TraceArgs):
        frame, _, _ = trace_args
        module_name = frame.f_globals.get('__name__')
        if module_name is None:
            return
        if module_name in self._modules_to_trace:
            return
        self._modules_to_trace.add(module_name)
        msg = f'{self.__class__.__name__}: added {module_name!r}'
        self._logger.info(msg)

    def _to_trace(self, trace_args: TraceArgs) -> bool:
        frame, _, _ = trace_args
        module_name = frame.f_globals.get('__name__')
        return match_any(module_name, self._modules_to_trace)



from nextline.spec import hookimpl
from nextline.types import RunNo
from nextline.utils.pubsub.broker import PubSub


class RunNoRegistrar:
    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_initialize_run(self, run_no: RunNo) -> None:
        await self._registry.publish('run_no', run_no)



from __future__ import annotations

from asyncio import Task
from threading import Thread
from types import FrameType
from typing import TYPE_CHECKING, Collection, Generator, Optional

import apluggy as pluggy
from apluggy import PluginManager, contextmanager

from nextline.spawned.types import QueueIn, QueueOut
from nextline.types import PromptNo, RunNo, TaskNo, ThreadNo, TraceNo

from .types import TraceArgs

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401


PROJECT_NAME = 'nextline_spawned'


hookspec = pluggy.HookspecMarker(PROJECT_NAME)
hookimpl = pluggy.HookimplMarker(PROJECT_NAME)


@hookspec
def init(
    hook: PluginManager,
    run_no: RunNo,
    modules_to_skip: Collection[str],
    queue_in: QueueIn,
    queue_out: QueueOut,
) -> None:
    pass


@hookspec
@contextmanager
def context():
    pass


@hookspec(firstresult=True)
def global_trace_func(frame: FrameType, event, arg) -> Optional[TraceFunc]:
    pass


@hookspec(firstresult=True)
def filter(trace_args: TraceArgs) -> bool | None:
    '''True to reject, False to accept, None to pass to the next hook implementation.

    Accepted if no hook implementation returns True or False.
    '''
    pass


@hookspec
def filtered(trace_args: TraceArgs) -> None:
    pass


@hookspec(firstresult=True)
def local_trace_func(frame: FrameType, event, arg) -> Optional[TraceFunc]:
    pass


@hookspec(firstresult=True)
def create_local_trace_func() -> Optional[TraceFunc]:
    pass


@hookspec
def on_start_task_or_thread() -> None:
    pass


@hookspec
def on_end_task_or_thread(task_or_thread: Task | Thread):
    pass


@hookspec
def on_start_trace(trace_no: TraceNo) -> None:
    pass


@hookspec(firstresult=True)
def current_thread_no() -> ThreadNo | None:
    pass


@hookspec(firstresult=True)
def current_task_no() -> TaskNo | None:
    pass


@hookspec(firstresult=True)
def current_trace_no() -> TraceNo | None:
    pass


@hookspec
def on_end_trace(trace_no: TraceNo) -> None:
    pass


@hookspec
@contextmanager
def on_trace_call(trace_args: TraceArgs):
    pass


@hookspec(firstresult=True)
def is_on_trace_call() -> Optional[bool]:
    pass


@hookspec(firstresult=True)
def current_trace_args() -> Optional[TraceArgs]:
    pass


@hookspec
@contextmanager
def on_cmdloop():
    pass


@hookspec
@contextmanager
def on_prompt(prompt_no: PromptNo, text: str) -> Generator[None, str, None]:
    # Receive the command by gen.send().
    command = yield  # noqa: F841
    yield


@hookspec(firstresult=True)
def prompt(prompt_no: PromptNo, text: str) -> Optional[str]:
    pass


@hookspec
def on_write_stdout(trace_no: TraceNo, line: str) -> None:
    pass



__all__ = [
    'PromptInfoRegistrar',
    'RunInfoRegistrar',
    'RunNoRegistrar',
    'ScriptRegistrar',
    'StateNameRegistrar',
    'StdoutRegistrar',
    'TraceInfoRegistrar',
    'TraceNumbersRegistrar',
]

from .registrars import (
    PromptInfoRegistrar,
    RunInfoRegistrar,
    RunNoRegistrar,
    ScriptRegistrar,
    StateNameRegistrar,
    StdoutRegistrar,
    TraceInfoRegistrar,
    TraceNumbersRegistrar,
)



from nextline.spec import hookimpl
from nextline.utils.pubsub.broker import PubSub


class StateNameRegistrar:
    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_change_state(self, state_name: str) -> None:
        await self._registry.publish('state_name', state_name)



import apluggy as pluggy
from apluggy import PluginManager

from nextline import spawned
from nextline.types import RunNo
from nextline.utils.pubsub.broker import PubSub

PROJECT_NAME = 'nextline_callback'


hookspec = pluggy.HookspecMarker(PROJECT_NAME)
hookimpl = pluggy.HookimplMarker(PROJECT_NAME)


@hookspec
def init(hook: PluginManager, registry: PubSub) -> None:
    pass


@hookspec
async def start() -> None:
    pass


@hookspec
async def close(exc_type=None, exc_value=None, traceback=None) -> None:
    pass


@hookspec
async def on_change_state(state_name: str) -> None:
    pass


@hookspec
async def on_change_script(script: str, filename: str) -> None:
    pass


@hookspec
async def on_initialize_run(run_no: RunNo) -> None:
    pass


@hookspec
async def on_start_run() -> None:
    pass


@hookspec
async def on_end_run(run_result: spawned.RunResult) -> None:
    pass


@hookspec
async def on_start_trace(event: spawned.OnStartTrace) -> None:
    pass


@hookspec
async def on_end_trace(event: spawned.OnEndTrace) -> None:
    pass


@hookspec
async def on_start_trace_call(event: spawned.OnStartTraceCall) -> None:
    pass


@hookspec
async def on_end_trace_call(event: spawned.OnEndTraceCall) -> None:
    pass


@hookspec
async def on_start_cmdloop(event: spawned.OnStartCmdloop) -> None:
    pass


@hookspec
async def on_end_cmdloop(event: spawned.OnEndCmdloop) -> None:
    pass


@hookspec
async def on_start_prompt(event: spawned.OnStartPrompt) -> None:
    pass


@hookspec
async def on_end_prompt(event: spawned.OnEndPrompt) -> None:
    pass


@hookspec
async def on_write_stdout(event: spawned.OnWriteStdout) -> None:
    pass



from contextlib import contextmanager
from logging import getLogger

from apluggy import PluginManager

from nextline.spawned.call import sys_trace
from nextline.spawned.types import QueueIn, QueueOut
from nextline.types import RunNo

from . import plugins, spec

MODULES_TO_SKIP = {
    'multiprocessing.*',
    'threading',
    'queue',
    'importlib',
    'asyncio.*',
    'codec',
    'concurrent.futures.*',
    'selectors',
    'weakref',
    '_weakrefset',
    'socket',
    'logging',
    'os',
    'collections.*',
    'importlib.*',
    'pathlib',
    'typing',
    'posixpath',
    'fnmatch',
    '_pytest.*',
    'apluggy.*',
    'pluggy.*',
    sys_trace.__module__,  # skip the 1st line of the finally clause in sys_trace()
    contextmanager.__module__,  # to skip contextlib.__exit__() in sys_trace()
}


def build_hook(
    run_no: RunNo,
    queue_in: QueueIn,
    queue_out: QueueOut,
) -> PluginManager:
    '''Return a plugin manager with the plugins registered.'''

    hook = PluginManager(spec.PROJECT_NAME)
    hook.add_hookspecs(spec)

    hook.register(plugins.Repeater)
    hook.register(plugins.PeekStdout)
    hook.register(plugins.Prompt)
    hook.register(plugins.PdbInstanceFactory)
    hook.register(plugins.TraceCallHandler)
    hook.register(plugins.LocalTraceFunc)
    hook.register(plugins.TaskOrThreadToTraceMapper)
    hook.register(plugins.TaskAndThreadKeeper)
    hook.register(plugins.FilerByModule)
    hook.register(plugins.FilterLambda)
    hook.register(plugins.FilterByModuleName)
    hook.register(plugins.GlobalTraceFunc)

    logger = getLogger(__name__)
    plugin_names = (f'{n!r}' for n, p in hook.list_name_plugin() if p)
    msg = f'Loaded plugins: {",".join(plugin_names)}.'
    logger.info(msg)

    hook.hook.init(
        hook=hook,
        run_no=run_no,
        modules_to_skip=MODULES_TO_SKIP,
        queue_in=queue_in,
        queue_out=queue_out,
    )

    return hook



from apluggy import PluginManager

from . import plugins, spec


def build_hook() -> PluginManager:
    hook = PluginManager(spec.PROJECT_NAME)
    hook.add_hookspecs(spec)

    hook.register(plugins.StdoutRegistrar)
    hook.register(plugins.PromptInfoRegistrar)
    hook.register(plugins.TraceInfoRegistrar)
    hook.register(plugins.TraceNumbersRegistrar)
    hook.register(plugins.RunInfoRegistrar)
    hook.register(plugins.RunNoRegistrar)
    hook.register(plugins.StateNameRegistrar)
    hook.register(plugins.ScriptRegistrar)

    return hook



from __future__ import annotations

import threading
from asyncio import Task
from logging import getLogger
from threading import Thread
from types import FrameType
from typing import TYPE_CHECKING, Iterator, MutableMapping, Optional
from weakref import WeakKeyDictionary, WeakSet

from apluggy import PluginManager, contextmanager

from nextline.count import TraceNoCounter
from nextline.spawned.plugin.spec import hookimpl
from nextline.types import TaskNo, ThreadNo, TraceNo
from nextline.utils import (
    ThreadTaskDoneCallback,
    ThreadTaskIdComposer,
    current_task_or_thread,
)

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401


class GlobalTraceFunc:
    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    def global_trace_func(self, frame: FrameType, event, arg) -> Optional[TraceFunc]:
        if self._hook.hook.filter(trace_args=(frame, event, arg)):
            return None
        self._hook.hook.filtered(trace_args=(frame, event, arg))
        return self._hook.hook.local_trace_func(frame=frame, event=event, arg=arg)


class TaskAndThreadKeeper:
    def __init__(self) -> None:
        self._set: WeakSet[Task | Thread] = WeakSet()
        self._counter = ThreadTaskIdComposer()
        self._main_thread: Optional[Thread] = None
        self._to_end: Optional[Thread] = None
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    @contextmanager
    def context(self) -> Iterator[None]:
        self._callback = ThreadTaskDoneCallback(done=self._on_end)
        self._main_thread = threading.current_thread()
        try:
            yield
        finally:
            self._callback.close()
            self._to_end and self._on_end(self._to_end)

    @hookimpl
    def filtered(self) -> None:
        current = current_task_or_thread()
        if current not in self._set:
            self._on_start(current)
            self._set.add(current)

    def _on_start(self, current: Task | Thread) -> None:
        self._logger.info(f'{self.__class__.__name__}._on_start: {current}')
        if current is self._main_thread:
            self._to_end = self._main_thread
        else:
            self._callback.register(current)
        self._counter()  # increment the counter
        self._hook.hook.on_start_task_or_thread()

    def _on_end(self, ending: Task | Thread):
        # The "ending" is not the "current" unless it is the main thread.
        self._logger.info(f'{self.__class__.__name__}._on_end: {ending}')
        self._hook.hook.on_end_task_or_thread(task_or_thread=ending)

    @hookimpl
    def current_thread_no(self) -> ThreadNo:
        return self._counter().thread_no

    @hookimpl
    def current_task_no(self) -> Optional[TaskNo]:
        return self._counter().task_no


class TaskOrThreadToTraceMapper:
    def __init__(self) -> None:
        self._map: MutableMapping[Task | Thread, TraceNo] = WeakKeyDictionary()
        self._counter = TraceNoCounter(1)
        self._logger = getLogger(__name__)

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    def on_start_task_or_thread(self) -> None:
        trace_no = self._counter()
        self._logger.info(f'{self.__class__.__name__} start: trace_no={trace_no}')
        self._map[current_task_or_thread()] = trace_no
        self._hook.hook.on_start_trace(trace_no=trace_no)

    @hookimpl
    def on_end_task_or_thread(self, task_or_thread: Task | Thread):
        trace_no = self._map[task_or_thread]
        self._hook.hook.on_end_trace(trace_no=trace_no)
        self._logger.info(f'{self.__class__.__name__} end: trace_no={trace_no}')

    @hookimpl
    def current_trace_no(self) -> Optional[TraceNo]:
        return self._map.get(current_task_or_thread())



import datetime
from typing import Generator

from apluggy import PluginManager, contextmanager

from nextline.spawned.events import (
    OnEndCmdloop,
    OnEndPrompt,
    OnEndTrace,
    OnEndTraceCall,
    OnStartCmdloop,
    OnStartPrompt,
    OnStartTrace,
    OnStartTraceCall,
    OnWriteStdout,
)
from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.plugin.types import TraceArgs
from nextline.spawned.types import QueueOut
from nextline.spawned.utils import to_canonic_path
from nextline.types import PromptNo, TraceNo


class Repeater:
    @hookimpl
    def init(self, hook: PluginManager, queue_out: QueueOut):
        self._hook = hook
        self._queue_out = queue_out

    @hookimpl
    def on_start_trace(self, trace_no: TraceNo):
        assert trace_no == self._hook.hook.current_trace_no()
        started_at = datetime.datetime.utcnow()
        thread_no = self._hook.hook.current_thread_no()
        task_no = self._hook.hook.current_task_no()
        event = OnStartTrace(
            started_at=started_at,
            trace_no=trace_no,
            thread_no=thread_no,
            task_no=task_no,
        )
        self._queue_out.put(event)

    @hookimpl
    def on_end_trace(self, trace_no: TraceNo) -> None:
        ended_at = datetime.datetime.utcnow()
        event = OnEndTrace(ended_at=ended_at, trace_no=trace_no)
        self._queue_out.put(event)

    @hookimpl
    @contextmanager
    def on_trace_call(self, trace_args: TraceArgs):
        started_at = datetime.datetime.utcnow()
        trace_no = self._hook.hook.current_trace_no()
        frame, call_event, call_arg = trace_args
        file_name = to_canonic_path(frame.f_code.co_filename)
        line_no = frame.f_lineno
        event_start = OnStartTraceCall(
            started_at=started_at,
            trace_no=trace_no,
            file_name=file_name,
            line_no=line_no,
            frame_object_id=id(frame),
            call_event=call_event,
        )
        self._queue_out.put(event_start)

        try:
            yield
        finally:
            ended_at = datetime.datetime.utcnow()
            event_end = OnEndTraceCall(ended_at=ended_at, trace_no=trace_no)
            self._queue_out.put(event_end)

    @hookimpl
    @contextmanager
    def on_cmdloop(self) -> Generator[None, None, None]:
        started_at = datetime.datetime.utcnow()
        trace_no = self._hook.hook.current_trace_no()
        event_start = OnStartCmdloop(started_at=started_at, trace_no=trace_no)
        self._queue_out.put(event_start)

        try:
            yield
        finally:
            ended_at = datetime.datetime.utcnow()
            event_end = OnEndCmdloop(ended_at=ended_at, trace_no=trace_no)
            self._queue_out.put(event_end)

    @hookimpl
    @contextmanager
    def on_prompt(self, prompt_no: PromptNo, text: str) -> Generator[None, str, None]:
        started_at = datetime.datetime.utcnow()
        trace_no = self._hook.hook.current_trace_no()
        event_start = OnStartPrompt(
            started_at=started_at,
            trace_no=trace_no,
            prompt_no=prompt_no,
            prompt_text=text,
        )
        self._queue_out.put(event_start)

        command = ''

        try:
            command = yield
            yield
        finally:
            ended_at = datetime.datetime.utcnow()
            event_end = OnEndPrompt(
                ended_at=ended_at,
                trace_no=trace_no,
                prompt_no=prompt_no,
                command=command,
            )
            self._queue_out.put(event_end)

    @hookimpl
    def on_write_stdout(self, trace_no: TraceNo, line: str):
        written_at = datetime.datetime.utcnow()
        trace_no = self._hook.hook.current_trace_no()
        event = OnWriteStdout(
            written_at=written_at,
            trace_no=trace_no,
            text=line,
        )
        self._queue_out.put(event)



'''Code used in the sub-processes in which the Nextline user code is run.
'''
from __future__ import annotations

__all__ = [
    'PdbCommand',
    'Event',
    'OnEndCmdloop',
    'OnEndPrompt',
    'OnEndTrace',
    'OnEndTraceCall',
    'OnStartCmdloop',
    'OnStartPrompt',
    'OnStartTrace',
    'OnStartTraceCall',
    'OnWriteStdout',
    'QueueIn',
    'QueueOut',
    'RunArg',
    'RunResult',
    'set_queues',
    'main',
]


from .commands import PdbCommand
from .events import (
    Event,
    OnEndCmdloop,
    OnEndPrompt,
    OnEndTrace,
    OnEndTraceCall,
    OnStartCmdloop,
    OnStartPrompt,
    OnStartTrace,
    OnStartTraceCall,
    OnWriteStdout,
)
from .run import run_
from .types import QueueIn, QueueOut, RunArg, RunResult

_queue_in: QueueIn | None = None
_queue_out: QueueOut | None = None


def set_queues(queue_in: QueueIn, queue_out: QueueOut) -> None:
    '''Initializer of ProcessPoolExecutor that receives the queues.'''
    global _queue_in, _queue_out
    _queue_in = queue_in
    _queue_out = queue_out


def main(run_arg: RunArg) -> RunResult:
    '''The function to be submitted to ProcessPoolExecutor.'''
    assert _queue_in
    assert _queue_out
    return run_(run_arg, _queue_in, _queue_out)



import dataclasses
import datetime
from typing import Optional

from nextline import spawned
from nextline.spec import hookimpl
from nextline.types import RunInfo, RunNo
from nextline.utils.pubsub.broker import PubSub


class RunInfoRegistrar:
    def __init__(self) -> None:
        self._run_no: Optional[RunNo] = None
        self._script: Optional[str] = None
        self._run_info: Optional[RunInfo] = None

    @hookimpl
    def init(self, registry: PubSub) -> None:
        self._registry = registry

    @hookimpl
    async def on_change_script(self, script: str) -> None:
        self._script = script

    @hookimpl
    async def on_initialize_run(self, run_no: RunNo) -> None:
        self._run_no = run_no
        self._run_info = RunInfo(
            run_no=run_no, state='initialized', script=self._script
        )
        await self._registry.publish('run_info', self._run_info)

    @hookimpl
    async def on_start_run(self) -> None:
        assert self._run_info is not None
        self._run_info = dataclasses.replace(
            self._run_info,
            state='running',
            started_at=datetime.datetime.utcnow(),
        )
        await self._registry.publish('run_info', self._run_info)

    @hookimpl
    async def on_end_run(self, run_result: spawned.RunResult) -> None:
        assert self._run_info is not None

        self._run_info = dataclasses.replace(
            self._run_info,
            state='finished',
            result=run_result.fmt_ret,
            exception=run_result.fmt_exc,
            ended_at=datetime.datetime.utcnow(),
        )
        await self._registry.publish('run_info', self._run_info)

        self._run_info = None
        self._run_no = None



from __future__ import annotations

import sys
from pathlib import Path
from types import CodeType
from typing import Any, Callable

from . import script
from .call import sys_trace
from .plugin import build_hook
from .trace import TraceFunc
from .types import QueueIn, QueueOut, RunArg, RunResult
from .utils import WithContext, to_canonic_path


def run_(run_arg: RunArg, queue_in: QueueIn, queue_out: QueueOut) -> RunResult:

    result = RunResult(ret=None, exc=None)

    try:
        func = _compose_callable(run_arg)
    except BaseException as e:
        result.exc = e
        return result

    run_no = run_arg.run_no
    hook = build_hook(run_no=run_no, queue_in=queue_in, queue_out=queue_out)

    with TraceFunc(hook=hook) as trace:
        with sys_trace(trace_func=trace):
            try:
                result.ret = func()
            except BaseException as e:
                result.exc = e

    # NOTE: How to print the exception in the same way as the interpreter.
    # import traceback
    # traceback.print_exception(type(exc), exc, exc.__traceback__)

    exc = result.exc
    if exc and exc.__traceback__:
        # remove this frame from the traceback.
        # Note: exc.__traceback__ is sys._getframe()
        exc.__traceback__ = exc.__traceback__.tb_next

    if exc and exc.__traceback__ and isinstance(exc, KeyboardInterrupt):
        tb = exc.__traceback__
        while tb.tb_next:
            module = tb.tb_next.tb_frame.f_globals.get('__name__')
            if module == WithContext.__module__:
                tb.tb_next = None
                break
            tb = tb.tb_next

    return result


def _compose_callable(run_arg: RunArg) -> Callable[[], Any]:
    # TODO: Rewrite with a match statement for Python 3.10
    statement = run_arg.statement
    filename = run_arg.filename

    if isinstance(statement, str):
        if (path := Path(to_canonic_path(statement))).is_file():
            statement = path

    if isinstance(statement, str):
        assert filename is not None
        code = compile(statement, filename, 'exec')
        return script.compose(code)
    elif isinstance(statement, Path):
        return _from_path(statement)
    elif isinstance(statement, CodeType):
        return script.compose(statement)
    elif callable(statement):
        return statement
    else:
        raise TypeError(f'statement: {statement!r}')


def _from_path(path: Path) -> Callable[[], Any]:
    # Read as a str and compile it as Pdb does.
    # https://github.com/python/cpython/blob/v3.10.10/Lib/pdb.py#L1568-L1592
    statement = path.read_text()
    code = compile(statement, str(path), 'exec')
    sys.path.insert(0, str(path.parent))
    return script.compose(code)



from __future__ import annotations

from collections import defaultdict
from types import FrameType
from typing import TYPE_CHECKING, Callable, DefaultDict, Dict, Optional, Set

from apluggy import PluginManager, contextmanager
from exceptiongroup import BaseExceptionGroup, catch

from nextline.spawned.plugin.spec import hookimpl
from nextline.spawned.plugin.types import TraceArgs
from nextline.spawned.utils import WithContext
from nextline.types import TraceNo

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401


class LocalTraceFunc:
    '''A plugin that executes local trace functions.

    It calls different trace functions for each trace number. If a trace
    function doesn't exist for a trace number, this plugin creates a new one by
    calling the hook `create_local_trace_func`.

    A trace number is assigned to each async task or thread by another plugin.
    '''

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook
        factory = Factory(hook)
        self._map: DefaultDict[TraceNo, TraceFunc] = defaultdict(factory)

    @hookimpl
    def local_trace_func(self, frame: FrameType, event, arg) -> Optional[TraceFunc]:
        trace_no = self._hook.hook.current_trace_no()
        local_trace_func = self._map[trace_no]
        return local_trace_func(frame, event, arg)


def Factory(hook: PluginManager) -> Callable[[], TraceFunc]:
    '''Return a function that creates a local trace function.'''

    def _factory() -> TraceFunc:
        trace = hook.hook.create_local_trace_func()

        @contextmanager
        def _context(frame, event, arg):
            '''A "with" block in which "trace" is called.'''

            keyboard_interrupt_raised = False

            def _keyboard_interrupt(exc: BaseExceptionGroup) -> None:
                nonlocal keyboard_interrupt_raised
                keyboard_interrupt_raised = True

            with hook.with_.on_trace_call(trace_args=(frame, event, arg)):
                with catch({KeyboardInterrupt: _keyboard_interrupt}):
                    # TODO: Using exceptiongroup.catch() for Python 3.8 and 3.9.
                    #       Rewrite with except* for Python 3.11.
                    #       https://pypi.org/project/exceptiongroup/

                    yield

            if keyboard_interrupt_raised:
                # Reraise after the "with" block so that gen.throw() is not called.
                raise KeyboardInterrupt

        return WithContext(trace, context=_context)

    return _factory


class TraceCallHandler:
    '''A plugin that keeps the trace call arguments during trace calls.

    This plugin collect the trace call arguments when the context manager hook
    `on_trace_call` is entered. It responds to the first result only hooks
    `is_on_trace_call` and `current_trace_args`.
    '''

    def __init__(self) -> None:
        self._trace_args_map: Dict[TraceNo, TraceArgs] = {}
        self._traces_on_call: Set[TraceNo] = set()

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._hook = hook

    @hookimpl
    @contextmanager
    def on_trace_call(self, trace_args: TraceArgs):
        trace_no = self._hook.hook.current_trace_no()
        self._traces_on_call.add(trace_no)
        self._trace_args_map[trace_no] = trace_args
        try:
            yield
        finally:
            self._traces_on_call.remove(trace_no)
            del self._trace_args_map[trace_no]

    @hookimpl
    def is_on_trace_call(self) -> Optional[bool]:
        trace_no = self._hook.hook.current_trace_no()
        return trace_no in self._traces_on_call

    @hookimpl
    def current_trace_args(self) -> Optional[TraceArgs]:
        trace_no = self._hook.hook.current_trace_no()
        return self._trace_args_map.get(trace_no)



from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING

from apluggy import PluginManager

from nextline.count import PromptNoCounter
from nextline.spawned.exc import NotOnTraceCall
from nextline.spawned.plugin.spec import hookimpl

from .custom import CustomizedPdb
from .stream import PromptFuncType, StdInOut

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401


class PdbInstanceFactory:
    '''A plugin that creates local trace functions.

    Each time the first result only hook `create_local_trace_func` is called,
    this plugin creates a new instance of the customized Pdb class and returns
    its trace function.

    The hook `create_local_trace_func` is called for each async task or thread.
    '''

    @hookimpl
    def init(self, hook: PluginManager) -> None:
        self._factory = Factory(hook=hook)

    @hookimpl
    def create_local_trace_func(self) -> TraceFunc:
        return self._factory()


def Factory(hook: PluginManager):
    cmdloop_hook = CmdloopHook(hook=hook)
    prompt_func = PromptFunc(hook=hook)

    def _factory():
        stdio = StdInOut(prompt_func=prompt_func)
        pdb = CustomizedPdb(
            cmdloop_hook=cmdloop_hook,
            stdin=stdio,
            stdout=stdio,
        )
        stdio.prompt_end = pdb.prompt
        return pdb.trace_dispatch

    return _factory


def CmdloopHook(hook: PluginManager):
    '''Return a context manager in which Pdb.cmdloop() is called.'''

    def cmdloop():
        if not hook.hook.is_on_trace_call():
            raise NotOnTraceCall
        return hook.with_.on_cmdloop()

    return cmdloop


def PromptFunc(hook: PluginManager) -> PromptFuncType:
    '''Return a function that responds to Pdb prompts with user commands.'''

    counter = PromptNoCounter(1)
    logger = getLogger(__name__)

    def _prompt_func(text: str) -> str:
        prompt_no = counter()
        logger.debug(f'PromptNo: {prompt_no}')
        with (context := hook.with_.on_prompt(prompt_no=prompt_no, text=text)):
            command = hook.hook.prompt(prompt_no=prompt_no, text=text)
            if command is None:
                logger.warning(f'command is None: {command!r}')
            context.gen.send(command)
        return command

    return _prompt_func



from __future__ import annotations

import sys
import threading
from contextlib import contextmanager
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from sys import TraceFunction as TraceFunc  # type: ignore  # noqa: F401


@contextmanager
def sys_trace(trace_func: TraceFunc, thread: Optional[bool] = True):
    '''Trace callables in the context and all threads created during the context.

    Notes
    -----
    All new threads created during the context are traced regardless of whether
    the threads are created by code in the context. If the thread option is
    false, no new threads will be traced.

    Example:

    Define a trace function.
    >>> def trace_func(frame, event, arg):
    ...     # Print the module name, function name, and event.
    ...     module_name = frame.f_globals.get('__name__')
    ...     func_name = frame.f_code.co_name
    ...     print(f'trace_func(): {module_name}.{func_name}() # {event}')
    ...
    ...     # Return the trace function only if the function name is 'callable'.
    ...     if func_name == 'callable':
    ...         return trace_func
    ...     return None

    Define a callable to trace.
    >>> def callable():
    ...     print('callable()')

    Trace the callable.
    >>> with sys_trace(trace_func):
    ...     callable()
    trace_func(): nextline.spawned.call.callable() # call
    trace_func(): nextline.spawned.call.callable() # line
    callable()
    trace_func(): nextline.spawned.call.callable() # return
    trace_func(): contextlib.__exit__() # call
    trace_func(): nextline.spawned.call.sys_trace() # call


    The callable is traced.

    However, as can be seen in the output, after the callable returns, the
    trace function continues to be called until sys.settrace(None) is called in
    the finally clause.

    '''
    trace_func_org = sys.gettrace()

    if thread:
        threading.settrace(trace_func)

    sys.settrace(trace_func)

    try:
        yield
    finally:
        sys.settrace(trace_func_org)
        if thread:
            threading.settrace(trace_func_org)  # type: ignore



from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Iterator, Optional

from apluggy import PluginManager, contextmanager

from .plugin import spec

if TYPE_CHECKING:
    from sys import TraceFunction  # type: ignore  # noqa: F401


assert spec.PROJECT_NAME  # used in the doctest


@contextmanager
def TraceFunc(hook: PluginManager) -> Iterator[TraceFunction]:
    '''Yield the trace function of Nextline to be set by sys.settrace().

    Python docs: sys.settrace():
    https://docs.python.org/3/library/sys.html#sys.settrace

    This yielded function only calls a hook function. The actual trace function
    is implemented in plugins.

    Example:

    Save the original trace function.
    >>> import sys
    >>> trace_org = sys.gettrace()

    Initialize a plugin manager of apluggy.
    >>> hook = PluginManager(spec.PROJECT_NAME)
    >>> hook.add_hookspecs(spec)

    >>> # In practice, plugins are registered here.

    Enter the "with" block:
    >>> with TraceFunc(hook) as trace_func:
    ...
    ...     # Set the trace function.
    ...     sys.settrace(trace_func)
    ...
    ...     # run the Nextline client function here
    ...
    ...     # Unset the trace function.
    ...     sys.settrace(trace_org)

    '''

    logger = getLogger(__name__)

    def _trace_func(frame, event, arg) -> Optional[TraceFunction]:
        try:
            return hook.hook.global_trace_func(frame=frame, event=event, arg=arg)
        except BaseException:
            logger.exception('')
            raise

    with hook.with_.context():
        logger.info(f'{TraceFunc.__name__}: started the hook')
        try:
            yield _trace_func
        finally:
            logger.info(f'{TraceFunc.__name__}: closing the hook')



from __future__ import annotations

import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from logging import getLogger
from typing import Any, AsyncIterator, Callable, Optional, Tuple

from apluggy import PluginManager, asynccontextmanager
from tblib import pickling_support

from . import spawned
from .count import RunNoCounter
from .hook import build_hook
from .monitor import Monitor
from .spawned import PdbCommand, QueueIn, QueueOut, RunArg, RunResult
from .types import PromptNo, RunNo, TraceNo
from .utils import MultiprocessingLogging, PubSub, Result, Running, run_in_process

pickling_support.install()

SCRIPT_FILE_NAME = "<string>"


def _call_all(*funcs) -> None:
    '''Execute callables and ignore return values.

    Used to call multiple initializers in ProcessPoolExecutor.
    '''
    for func in funcs:
        func()


@asynccontextmanager
async def run_with_resource(
    hook: PluginManager, run_arg: RunArg
) -> AsyncIterator[Tuple[Running[RunResult], Callable[[str, int, int], None]]]:
    mp_context = mp.get_context('spawn')
    queue_in: QueueIn = mp_context.Queue()
    queue_out: QueueOut = mp_context.Queue()
    monitor = Monitor(hook, queue_out)
    async with MultiprocessingLogging(mp_context=mp_context) as mp_logging:
        async with monitor:
            initializer = partial(
                _call_all,
                mp_logging.initializer,
                partial(spawned.set_queues, queue_in, queue_out),
            )
            executor_factory = partial(
                ProcessPoolExecutor,
                max_workers=1,
                mp_context=mp_context,
                initializer=initializer,
            )
            func = partial(spawned.main, run_arg)
            running = await run_in_process(func, executor_factory)
            send_pdb_command = SendPdbCommand(queue_in)
            yield running, send_pdb_command


def SendPdbCommand(queue_in: QueueIn) -> Callable[[str, int, int], None]:
    def _send_pdb_command(command: str, prompt_no: int, trace_no: int) -> None:
        logger = getLogger(__name__)
        logger.debug(f'send_pdb_command({command!r}, {prompt_no!r}, {trace_no!r})')
        item = PdbCommand(
            trace_no=TraceNo(trace_no),
            prompt_no=PromptNo(prompt_no),
            command=command,
        )
        queue_in.put(item)

    return _send_pdb_command


class Context:
    def __init__(self, run_no_start_from: int, statement: str):
        self._hook = build_hook()
        self.registry = PubSub[Any, Any]()
        self._hook.hook.init(hook=self._hook, registry=self.registry)
        self._run_no_count = RunNoCounter(run_no_start_from)
        self._running: Optional[Running[RunResult]] = None
        self._send_pdb_command: Optional[Callable[[str, int, int], None]] = None
        self._run_arg = RunArg(
            run_no=RunNo(run_no_start_from - 1),
            statement=statement,
            filename=SCRIPT_FILE_NAME,
        )
        self._run_result: RunResult | None = None

    async def start(self) -> None:
        await self._hook.ahook.start()
        await self._hook.ahook.on_change_script(
            script=self._run_arg.statement, filename=self._run_arg.filename
        )

    async def state_change(self, state_name: str):
        await self._hook.ahook.on_change_state(state_name=state_name)

    async def shutdown(self) -> None:
        await self.registry.close()
        await self._hook.ahook.close()

    async def initialize(self) -> None:
        self._run_arg.run_no = self._run_no_count()
        self._run_result = None
        await self._hook.ahook.on_initialize_run(run_no=self._run_arg.run_no)

    async def reset(
        self,
        statement: Optional[str] = None,
        run_no_start_from: Optional[int] = None,
    ):
        if statement:
            self._run_arg.statement = statement
            await self._hook.ahook.on_change_script(
                script=statement, filename=self._run_arg.filename
            )
        if run_no_start_from is not None:
            self._run_no_count = RunNoCounter(run_no_start_from)

    @asynccontextmanager
    async def run(self) -> AsyncIterator[None]:
        try:
            async with run_with_resource(self._hook, self._run_arg) as (
                running,
                send_pdb_command,
            ):
                self._running = running
                self._send_pdb_command = send_pdb_command
                await self._hook.ahook.on_start_run()
                yield
                ret = await running
                self._running = None
                self._send_pdb_command = None
        finally:
            await self._finish(ret)

    async def _finish(self, ret: Result[RunResult]) -> None:
        self._run_result = ret.returned or RunResult(ret=None, exc=None)
        if ret.raised:
            logger = getLogger(__name__)
            logger.exception(ret.raised)
        await self._hook.ahook.on_end_run(run_result=self._run_result)

    def send_pdb_command(self, command: str, prompt_no: int, trace_no: int) -> None:
        if self._send_pdb_command:
            self._send_pdb_command(command, prompt_no, trace_no)

    def interrupt(self) -> None:
        if self._running:
            self._running.interrupt()

    def terminate(self) -> None:
        if self._running:
            self._running.terminate()

    def kill(self) -> None:
        if self._running:
            self._running.kill()

    def result(self) -> Any:
        assert self._run_result
        return self._run_result.result()

    def exception(self) -> Optional[BaseException]:
        assert self._run_result
        return self._run_result.exc

    async def close(self):
        pass
